#ifndef terminalstatus_h_included_p
#define terminalstatus_h_included_p

///@file
/// Declare class cxxtls::TerminalStatus

//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.


#include <memory>

namespace cxxtls
{
struct TerminalState;  ///<  Same as 'struct termios', but is a
                       /// work around for the fact that 
                       /// struct termios is defined in
                       /// header files that ALSO define annoying
                       /// macros

class TerminalStatus
    ///  A object which holds the contents of struct termios and which can be used
    ///  to read or write the termios from/to a file handle.
    ///
    /// Construct TerminalStatus objects from a file handle -- that is, let the constructor
    /// read the file handle handle to get its contents.  Then use the various  member 
    /// functions to modify the status  before you then use setStatus() to copy the status 
    /// to a file handle.  See setStatus(), setRaw(), setNoEcho(), etc.
{
    std::auto_ptr<TerminalState> override_termios_;

    bool failed_;

public:

    TerminalStatus(int fd=0);  ///<  Construct given 0 or
                               ///   the master file descriptor
                               ///   of a PsuedoTerminal

    TerminalStatus(TerminalStatus const &rhs);

    TerminalStatus &operator=(TerminalStatus const &rhs);
       ///<  Assign from another terminal status object


    void setRaw();   ///<  Modify *this so that it refers to a
                     ///   truly "raw" terminal interface.
                     ///
                     ///   Use this in preparation for calling
                     ///   setStatus(fd) on some file descriptor
                     ///   to force that file descriptor to 
                     ///   match the state defined in *this

    void setNoEcho();///<  Modify *this so that it no longer 
                     ///   has the echo property turned on.
                     ///   Typically you do this if you are
                     ///   pre-processing the input and are
                     ///   echoing it yourself.  
                     ///
                     ///   Use this in preparation for calling
                     ///   setStatus(fd) on some file descriptor
                     ///   to force that file descriptor to 
                     ///   match the state defined in *this

    void setSignal(bool enabled);  ///<  Modifiy *this such that ^C -> SIGINT or not, 
                                   ///   depending on @b enabled.
                                   ///
                                   ///   @param enabled  True implies that signals should
                                   ///                   be enabled.


    int setStatus(int fd);
       ///< force the terminal status on a file handle.  @return a negative number
       /// if the operation fails.  See errno

}; // struct TerminalStatus

} // namespace cxxtls

#endif
